@@ -195,8 +195,9 @@ class CXCircuitLayer:
              figsize: Optional[Tuple[int, int]] = None,
              zcolor: str = "#CCFFCC",
              xcolor: str = "#FF8888",
+             noshow: bool = False,
              **kwargs):
-        # pylint: disable = too-many-locals
+        # pylint: disable = too-many-locals, too-many-branches
         """
             Draws this CX circuit layer using NetworkX and Matplotlib.
 
@@ -243,9 +244,11 @@ class CXCircuitLayer:
         for ctrl, trgt in self.gates:
             edge_idx = edges.index(Coupling(ctrl, trgt))
             kwargs["edge_color"][edge_idx] = "#000000"
-        plt.figure(figsize=figsize)
+        if not noshow:
+            plt.figure(figsize=figsize)
         nx.draw_networkx(G, **kwargs)
-        plt.show()
+        if not noshow:
+            plt.show()
 
     def __irshift__(self, gates: Union[GateLike, Sequence[GateLike]]) -> "CXCircuitLayer":
         if (isinstance(gates, (list, tuple))
@@ -289,7 +292,8 @@ class CXCircuitLayer:
                 yield (fst, snd)
                 yield (snd, fst)
 
-CXCircuitLayerLike = Union[CXCircuitLayer, Sequence[GateLike]]
+CXCircuitLayerLike = Union[CXCircuitLayer, "CXCircuitLayerView", Sequence[GateLike]]
+CXCircuitLike = Union["CXCircuit", "CXCircuitView", Sequence[CXCircuitLayerLike]]
 
 class CXCircuit(Sequence[CXCircuitLayer]):
     """
@@ -351,10 +355,24 @@ class CXCircuit(Sequence[CXCircuitLayer]):
             Draws this CX circuit using NetworkX and Matplotlib.
             Keyword arguments `kwargs` are those of `networkx.draw_networkx`.
         """
+        try:
+            # pylint: disable = import-outside-toplevel
+            import matplotlib.pyplot as plt # type: ignore
+        except ModuleNotFoundError as _:
+            raise ModuleNotFoundError("You must install the 'matplotlib' library.")
+        large_figsize = None
+        if figsize is not None:
+            w, h = figsize
+            large_figsize = (w*len(self), h)
+        plt.figure(figsize=large_figsize)
+        # plt.tight_layout()
+        plt.subplots_adjust(wspace=0)
         for layer_idx, layer in enumerate(self):
-            print(f"Layer {layer_idx}:")
-            layer.draw(layout=layout, figsize=figsize,
+            # print(f"Layer {layer_idx}:")
+            plt.subplot(1, len(self), layer_idx+1, title=f"Layer {layer_idx}")
+            layer.draw(layout=layout, figsize=figsize, noshow=True,
                        zcolor=zcolor, xcolor=xcolor, **kwargs)
+        plt.show()
 
     @overload
     def __getitem__(self, layer_idx: int) -> CXCircuitLayer:
@@ -374,9 +392,11 @@ class CXCircuit(Sequence[CXCircuitLayer]):
         return iter(self._layers)
 
     def __irshift__(self, layers: Union[CXCircuitLayerLike,
-                                        Sequence[CXCircuitLayerLike]]) -> "CXCircuit":
-        if isinstance(layers, CXCircuitLayer):
+                                        CXCircuitLike]) -> "CXCircuit":
+        if isinstance(layers, (CXCircuitLayer, CXCircuitLayerView)):
             layers = [layers]
+        if isinstance(layers, (CXCircuit, CXCircuitView)):
+            layers = [layer.clone() for layer in layers]
         elif (isinstance(layers, Sequence)
               and all(isinstance(g, (list, tuple))
                       and all(isinstance(x, int) for x in g) # pylint: disable = C0330
@@ -385,13 +405,21 @@ class CXCircuit(Sequence[CXCircuitLayer]):
         if not isinstance(layers, Sequence):
             raise TypeError(f"Expected sequence of layers, found {layers}")
         for layer in layers:
-            if not isinstance(layer, CXCircuitLayer):
+            if isinstance(layer, CXCircuitLayerView):
+                layer = layer.clone()
+            elif not isinstance(layer, CXCircuitLayer):
                 if not isinstance(layer, Sequence):
                     raise TypeError(f"Expected a sequence of pairs of ints, found {layer}")
                 layer = CXCircuitLayer(self.topology, cast(Sequence[GateLike], layer))
             self._layers.append(layer)
         return self
 
+    def __rshift__(self, layers: Union[CXCircuitLayerLike,
+                                       CXCircuitLike]) -> "CXCircuit":
+        circ = CXCircuit(self.topology, [])
+        circ >>= layers
+        return circ
+
     def __eq__(self, other) -> bool:
         if self is other:
             return True
